@@ -1,19 +1,21 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import hashlib
+from collections import defaultdict
 
-from sql import Null
+from sql.functions import CharLength
 
 from trytond.i18n import gettext
-from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, Check
+from trytond.model import (
+    Check, Index, Model, ModelSQL, ModelView, Workflow, fields)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, Bool, If
-from trytond.transaction import Transaction
 from trytond.pool import Pool
-from trytond.wizard import Wizard, StateView, StateTransition, Button
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import grouped_slice, is_full_text, lstrip_wildcard
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 from .exceptions import (
-    InventoryValidationError, InventoryCountWarning, InventoryFutureWarning)
+    InventoryCountWarning, InventoryFutureWarning, InventoryValidationError)
 
 
 class Inventory(Workflow, ModelSQL, ModelView):
@@ -24,7 +26,6 @@ class Inventory(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Eval('state') != 'draft',
         }
-    _depends = ['state']
 
     number = fields.Char('Number', readonly=True,
         help="The main identifier for the inventory.")
@@ -33,45 +34,44 @@ class Inventory(Workflow, ModelSQL, ModelView):
         domain=[('type', '=', 'storage')], states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The location inventoried.")
     date = fields.Date('Date', required=True, states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The date of the stock count.")
     lines = fields.One2Many(
         'stock.inventory.line', 'inventory', 'Lines',
         states={
             'readonly': (_states['readonly'] | ~Eval('location')
                 | ~Eval('date')),
-            },
-        depends=['location', 'date'] + _depends)
+            })
     empty_quantity = fields.Selection([
             (None, ""),
             ('keep', "Keep"),
             ('empty', "Empty"),
-            ], "Empty Quantity", states=_states, depends=_depends,
+            ], "Empty Quantity", states=_states,
         help="How lines without a quantity are handled.")
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The company the inventory is associated with.")
     state = fields.Selection([
             ('draft', "Draft"),
             ('done', "Done"),
             ('cancelled', "Cancelled"),
-            ], "State", readonly=True, select=True,
+            ], "State", readonly=True, sort=False,
         help="The current state of the inventory.")
 
-    del _states, _depends
+    del _states
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
         super(Inventory, cls).__setup__()
-        cls.create_date.select = True
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(t, (t.state, Index.Equality()), where=t.state == 'draft'))
         cls._order.insert(0, ('date', 'DESC'))
         cls._transitions |= set((
                 ('draft', 'done'),
@@ -90,7 +90,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
                     'readonly': Eval('state') != 'draft',
                     'depends': ['state'],
                     },
-                'count': {
+                'do_count': {
                     'readonly': Eval('state') != 'draft',
                     'depends': ['state'],
                     },
@@ -112,6 +112,11 @@ class Inventory(Workflow, ModelSQL, ModelView):
                 [sql_table.state], ['cancelled'],
                 where=sql_table.state == 'cancel'))
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @staticmethod
     def default_state():
         return 'draft'
@@ -125,6 +130,28 @@ class Inventory(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    def get_rec_name(self, name):
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        lang = Lang.get()
+        date = lang.strftime(self.date)
+        return f"[{self.number}] {self.location.rec_name} @ {date}"
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        number_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            number_value = lstrip_wildcard(operand)
+        return [bool_op,
+            ('number', operator, number_value, *extra),
+            ('location.rec_name', operator, operand, *extra),
+            ]
+
     @classmethod
     def view_attributes(cls):
         return super().view_attributes() + [
@@ -164,9 +191,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
             names = ', '.join(i.rec_name for i in future_inventories[:5])
             if len(future_inventories) > 5:
                 names + '...'
-            warning_name = (
-                '%s.date_future' % hashlib.md5(
-                    str(future_inventories).encode('utf-8')).hexdigest())
+            warning_name = Warning.format('date_future', future_inventories)
             if Warning.check(warning_name):
                 raise InventoryFutureWarning(warning_name,
                     gettext('stock.msg_inventory_date_in_the_future',
@@ -259,17 +284,21 @@ class Inventory(Workflow, ModelSQL, ModelView):
             if inventory.state == 'done':
                 continue
             # Compute product quantities
-            if fill:
-                product_ids = None
-            else:
-                product_ids = [l.product.id for l in inventory.lines]
             with Transaction().set_context(
                     company=inventory.company.id,
                     stock_date_end=inventory.date):
-                pbl = Product.products_by_location(
-                    [inventory.location.id],
-                    grouping=grouping,
-                    grouping_filter=(product_ids,))
+                if fill:
+                    pbl = Product.products_by_location(
+                        [inventory.location.id],
+                        grouping=grouping)
+                else:
+                    product_ids = [l.product.id for l in inventory.lines]
+                    pbl = defaultdict(int)
+                    for product_ids in grouped_slice(product_ids):
+                        pbl.update(Product.products_by_location(
+                                [inventory.location.id],
+                                grouping=grouping,
+                                grouping_filter=(list(product_ids),)))
 
             # Index some data
             product2type = {}
@@ -315,7 +344,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button_action('stock.wizard_inventory_count')
-    def count(cls, inventories):
+    def do_count(cls, inventories):
         cls.complete_lines(inventories)
 
 
@@ -325,26 +354,23 @@ class InventoryLine(ModelSQL, ModelView):
     _states = {
         'readonly': Eval('inventory_state') != 'draft',
         }
-    _depends = ['inventory_state']
 
     product = fields.Many2One('product.product', 'Product', required=True,
         domain=[
             ('type', '=', 'goods'),
-            ], states=_states, depends=_depends)
-    uom = fields.Function(fields.Many2One('product.uom', 'UOM',
-        help="The unit in which the quantity is specified."), 'get_uom')
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-            'get_unit_digits')
-    expected_quantity = fields.Float('Expected Quantity', required=True,
-        digits=(16, Eval('unit_digits', 2)), readonly=True,
+            ], states=_states)
+    unit = fields.Function(fields.Many2One(
+            'product.uom', "Unit",
+            help="The unit in which the quantity is specified."),
+        'get_unit')
+    expected_quantity = fields.Float(
+        "Expected Quantity", digits='unit', required=True, readonly=True,
         states={
             'invisible': Eval('id', -1) < 0,
         },
-        depends=['unit_digits'],
         help="The quantity the system calculated should be in the location.")
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
-        states=_states, depends=['unit_digits'] + _depends,
+    quantity = fields.Float(
+        "Actual Quantity", digits='unit', states=_states,
         help="The actual quantity found in the location.")
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
     inventory = fields.Many2One('stock.inventory', 'Inventory', required=True,
@@ -352,7 +378,6 @@ class InventoryLine(ModelSQL, ModelView):
         states={
             'readonly': _states['readonly'] & Bool(Eval('inventory')),
             },
-        depends=_depends,
         help="The inventory the line belongs to.")
     inventory_location = fields.Function(
         fields.Many2One('stock.location', "Location"),
@@ -364,7 +389,7 @@ class InventoryLine(ModelSQL, ModelView):
         searcher='search_inventory_date')
     inventory_state = fields.Function(
         fields.Selection('get_inventory_states', "Inventory State",
-        depends=['inventory']),
+        depends={'inventory'}),
         'on_change_with_inventory_state')
 
     @classmethod
@@ -378,53 +403,18 @@ class InventoryLine(ModelSQL, ModelView):
             ]
         cls._order.insert(0, ('product', 'ASC'))
 
-    @classmethod
-    def __register__(cls, module_name):
-        transaction = Transaction()
-        cursor = transaction.connection.cursor()
-        update = transaction.connection.cursor()
-        pool = Pool()
-        Move = pool.get('stock.move')
-        sql_table = cls.__table__()
-        move_table = Move.__table__()
-
-        super(InventoryLine, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-
-        # Migration from 3.0: use Move origin
-        if table.column_exist('move'):
-            cursor.execute(*sql_table.select(sql_table.id, sql_table.move,
-                    where=sql_table.move != Null))
-            for line_id, move_id in cursor:
-                update.execute(*move_table.update(
-                        columns=[move_table.origin],
-                        values=['%s,%s' % (cls.__name__, line_id)],
-                        where=move_table.id == move_id))
-            table.drop_column('move')
-
-        # Migration from 4.6: drop required on quantity
-        table.not_null_action('quantity', action='remove')
-
-    @staticmethod
-    def default_unit_digits():
-        return 2
-
     @staticmethod
     def default_expected_quantity():
         return 0.
 
     @fields.depends('product')
     def on_change_product(self):
-        self.unit_digits = 2
         if self.product:
-            self.uom = self.product.default_uom
-            self.unit_digits = self.product.default_uom.digits
+            self.unit = self.product.default_uom
 
     @fields.depends('inventory', '_parent_inventory.location')
     def on_change_with_inventory_location(self, name=None):
-        if self.inventory and self.inventory.location:
-            return self.inventory.location.id
+        return self.inventory.location if self.inventory else None
 
     @classmethod
     def search_inventory_location(cls, name, clause):
@@ -459,11 +449,8 @@ class InventoryLine(ModelSQL, ModelView):
     def search_rec_name(cls, name, clause):
         return [('product.rec_name',) + tuple(clause[1:])]
 
-    def get_uom(self, name):
-        return self.product.default_uom.id
-
-    def get_unit_digits(self, name):
-        return self.product.default_uom.digits
+    def get_unit(self, name):
+        return self.product.default_uom
 
     @property
     def unique_key(self):
@@ -488,7 +475,6 @@ class InventoryLine(ModelSQL, ModelView):
         '''
         pool = Pool()
         Move = pool.get('stock.move')
-        Uom = pool.get('product.uom')
 
         qty = self.quantity
         if qty is None:
@@ -501,9 +487,7 @@ class InventoryLine(ModelSQL, ModelView):
             else:
                 qty = 0.0
 
-        delta_qty = Uom.compute_qty(self.uom,
-            self.expected_quantity - qty,
-            self.uom)
+        delta_qty = self.unit.round(self.expected_quantity - qty)
         if delta_qty == 0.0:
             return
         from_location = self.inventory.location
@@ -522,7 +506,7 @@ class InventoryLine(ModelSQL, ModelView):
             to_location=to_location,
             quantity=delta_qty,
             product=self.product,
-            uom=self.uom,
+            unit=self.unit,
             company=self.inventory.company,
             effective_date=self.inventory.date,
             origin=self,
@@ -598,10 +582,9 @@ class Count(Wizard):
             line, = lines
         values['line'] = line.id
         values['product'] = line.product.id
-        values['uom'] = line.uom.id
-        values['unit_digits'] = line.unit_digits
-        if line.uom.rounding == 1:
-            values['quantity'] = 1
+        values['unit'] = line.unit.id
+        if line.unit.rounding == 1:
+            values['quantity'] = 1.
         return values
 
     def get_line_domain(self, inventory):
@@ -643,31 +626,18 @@ class CountSearch(ModelView):
             ('product.product', "Product"),
             ],
         required=True,
-        domain=[If(Eval('search_model') == 'product.product',
-                [
-                    ('type', '=', 'goods'),
-                    ('consumable', '=', False),
-                    ],
-                [])],
-        depends=['search_model'],
+        domain={
+            'product.product': [
+                ('type', '=', 'goods'),
+                ('consumable', '=', False),
+                ],
+            },
         help="The item that's counted.")
-    search_model = fields.Function(fields.Selection(
-        'get_search_models', "Search Model"),
-        'on_change_with_search_model')
 
     @classmethod
     def default_search(cls):
         return 'product.product,-1'
 
-    @classmethod
-    def get_search_models(cls):
-        return cls.fields_get(['search'])['search']['selection']
-
-    @fields.depends('search')
-    def on_change_with_search_model(self, name=None):
-        if self.search:
-            return self.search.__name__
-
 
 class CountQuantity(ModelView):
     "Stock Inventory Count"
@@ -676,20 +646,17 @@ class CountQuantity(ModelView):
     line = fields.Many2One(
         'stock.inventory.line', "Line", readonly=True, required=True)
     product = fields.Many2One('product.product', "Product", readonly=True)
-    uom = fields.Many2One('product.uom', "UOM", readonly=True,
+    unit = fields.Many2One(
+        'product.uom', "Unit", readonly=True,
         help="The unit in which the quantities are specified.")
     total_quantity = fields.Float(
-        "Total Quantity", digits=(16, Eval('unit_digits', 2)),
-        readonly=True, depends=['unit_digits'],
+        "Total Quantity", digits='unit', readonly=True,
         help="The total amount of the line counted so far.")
 
     quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
-        depends=['unit_digits'],
+        "Quantity", digits='unit', required=True,
         help="The quantity to add to the existing count.")
 
-    unit_digits = fields.Integer("Unit Digits", readonly=True)
-
     @fields.depends('quantity', 'line')
     def on_change_quantity(self):
         if self.line:
