@@ -1,12 +1,13 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from decimal import Decimal
-from itertools import tee, zip_longest
 
-from sql import Null, Literal, With
-from sql.aggregate import Sum, Min, Max
+from collections import defaultdict
+from decimal import Decimal
+
+from sql import Literal, Null, With
+from sql.aggregate import Max, Min, Sum
 from sql.conditionals import Case, Coalesce
-from sql.functions import CurrentTimestamp, DateTrunc, Power, Ceil, Log, Round
+from sql.functions import Ceil, CurrentTimestamp, DateTrunc, Log, Power, Round
 
 try:
     import pygal
@@ -15,50 +16,41 @@ except ImportError:
 from dateutil.relativedelta import relativedelta
 
 from trytond.i18n import lazy_gettext
-from trytond.model import ModelSQL, ModelView, fields
+from trytond.model import ModelSQL, ModelView, fields, sum_tree
+from trytond.modules.currency.fields import Monetary
 from trytond.pool import Pool
 from trytond.pyson import Eval, If
-from trytond.tools import grouped_slice, reduce_ids
+from trytond.tools import grouped_slice, pairwise_longest, reduce_ids
 from trytond.transaction import Transaction
 
 
-def pairwise(iterable):
-    a, b = tee(iterable)
-    next(b, None)
-    return zip_longest(a, b)
-
-
 class Abstract(ModelSQL, ModelView):
 
     company = fields.Many2One(
         'company.company', lazy_gettext('stock.msg_stock_reporting_company'))
-    cost = fields.Numeric(
+    cost = Monetary(
         lazy_gettext('stock.msg_stock_reporting_cost'),
-        digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
-    revenue = fields.Numeric(
+        currency='currency', digits='currency')
+    revenue = Monetary(
         lazy_gettext('stock.msg_stock_reporting_revenue'),
-        digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
-    profit = fields.Numeric(
+        currency='currency', digits='currency')
+    profit = Monetary(
         lazy_gettext('stock.msg_stock_reporting_profit'),
-        digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
+        currency='currency', digits='currency')
     margin = fields.Numeric(
         lazy_gettext('stock.msg_stock_reporting_margin'),
         digits=(14, 4),
         states={
             'invisible': ~Eval('margin'),
-            },
-        depends=['currency_digits'])
+            })
     margin_trend = fields.Function(fields.Char(
             lazy_gettext('stock.msg_stock_reporting_margin_trend')),
         'get_trend')
     time_series = None
 
-    currency = fields.Many2One('currency.currency', "Currency")
-    currency_digits = fields.Integer(
-        lazy_gettext('stock.msg_stock_reporting_currency_digits'))
+    currency = fields.Many2One(
+        'currency.currency',
+        lazy_gettext('stock.msg_stock_reporting_currency'))
 
     @classmethod
     def table_query(cls):
@@ -145,7 +137,6 @@ class Abstract(ModelSQL, ModelView):
             cls.margin.sql_cast(
                 Round(margin, cls.margin.digits[1])).as_('margin'),
             currency.id.as_('currency'),
-            currency.digits.as_('currency_digits'),
             ]
 
     @classmethod
@@ -213,7 +204,7 @@ class Abstract(ModelSQL, ModelView):
     @property
     def time_series_all(self):
         delta = self._period_delta()
-        for ts, next_ts in pairwise(self.time_series or []):
+        for ts, next_ts in pairwise_longest(self.time_series or []):
             yield ts
             if delta and next_ts:
                 date = ts.date + delta
@@ -286,15 +277,13 @@ class Context(ModelView):
             If(Eval('to_date') & Eval('from_date'),
                 ('from_date', '<=', Eval('to_date')),
                 ()),
-            ],
-        depends=['to_date'])
+            ])
     to_date = fields.Date("To Date",
         domain=[
             If(Eval('from_date') & Eval('to_date'),
                 ('to_date', '>=', Eval('from_date')),
                 ()),
-            ],
-        depends=['from_date'])
+            ])
     period = fields.Selection([
             ('year', "Year"),
             ('month', "Month"),
@@ -336,6 +325,28 @@ class Context(ModelView):
         return Transaction().context.get('include_lost', False)
 
 
+class Main(Abstract, ModelView):
+    "Stock Reporting Margin"
+    __name__ = 'stock.reporting.margin.main'
+
+    time_series = fields.Function(fields.One2Many(
+            'stock.reporting.margin.main.time_series', None, "Time Series"),
+        'get_time_series')
+
+    def get_rec_name(self, name):
+        return ''
+
+    def get_time_series(self, name):
+        pool = Pool()
+        Timeseries = pool.get('stock.reporting.margin.main.time_series')
+        return [t.id for t in Timeseries.search([])]
+
+
+class MainTimeseries(AbstractTimeseries, ModelView):
+    "Stock Reporting Margin"
+    __name__ = 'stock.reporting.margin.main.time_series'
+
+
 class ProductMixin:
     __slots__ = ()
 
@@ -344,20 +355,38 @@ class ProductMixin:
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     internal_quantity = fields.Float("Internal Quantity")
     quantity = fields.Function(fields.Float(
-            "Quantity", digits=(16, Eval('unit_digits', 2)),
-            depends=['unit_digits']), 'get_quantity')
-    unit = fields.Function(fields.Many2One('product.uom', "Unit"), 'get_unit')
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'get_unit_digits')
+            "Quantity", digits='unit'), 'get_quantity')
+    unit = fields.Many2One('product.uom', "Unit")
+
+    @classmethod
+    def _joins(cls):
+        pool = Pool()
+        Product = pool.get('product.product')
+        Template = pool.get('product.template')
+        from_item, tables, withs = super()._joins()
+        if 'move.product' not in tables:
+            product = Product.__table__()
+            tables['move.product'] = product
+            move = tables['move']
+            from_item = (from_item
+                .join(product, condition=move.product == product.id))
+        if 'move.product.template' not in tables:
+            template = Template.__table__()
+            tables['move.product.template'] = template
+            product = tables['move.product']
+            from_item = (from_item
+                .join(template, condition=product.template == template.id))
+        return from_item, tables, withs
 
     @classmethod
     def _columns(cls, tables, withs):
         move = tables['move']
         from_location = tables['move.from_location']
         to_location = tables['move.to_location']
+        template = tables['move.product.template']
         sign = Case(
             (from_location.type.in_(cls._to_location_types())
                 & to_location.type.in_(cls._from_location_types()),
@@ -366,25 +395,26 @@ class ProductMixin:
         return super()._columns(tables, withs) + [
             move.product.as_('product'),
             Sum(sign * move.internal_quantity).as_('internal_quantity'),
+            template.default_uom.as_('unit'),
             ]
 
     @classmethod
     def _group_by(cls, tables, withs):
         move = tables['move']
-        return super()._group_by(tables, withs) + [move.product]
+        template = tables['move.product.template']
+        return super()._group_by(tables, withs) + [
+            move.product, template.default_uom]
 
     def get_rec_name(self, name):
         return self.product.rec_name
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('product.rec_name', *clause[1:])]
+
     def get_quantity(self, name):
         return self.unit.round(self.internal_quantity)
 
-    def get_unit(self, name):
-        return self.product.default_uom.id
-
-    def get_unit_digits(self, name):
-        return self.product.default_uom.digits
-
 
 class Product(ProductMixin, Abstract, ModelView):
     "Stock Reporting Margin per Product"
@@ -469,6 +499,10 @@ class CategoryMixin:
     def get_rec_name(self, name):
         return self.category.rec_name if self.category else None
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('category.rec_name', *clause[1:])]
+
 
 class Category(CategoryMixin, Abstract, ModelView):
     "Stock Reporting Margin per Category"
@@ -498,31 +532,32 @@ class CategoryTree(ModelSQL, ModelView):
     "Stock Reporting Margin per Category"
     __name__ = 'stock.reporting.margin.category.tree'
 
-    name = fields.Function(fields.Char("Name"), 'get_name')
+    name = fields.Function(
+        fields.Char("Name"), 'get_name', searcher='search_name')
     parent = fields.Many2One('stock.reporting.margin.category.tree', "Parent")
     children = fields.One2Many(
         'stock.reporting.margin.category.tree', 'parent', "Children")
-    cost = fields.Function(fields.Numeric(
+    cost = fields.Function(Monetary(
             lazy_gettext('stock.msg_stock_reporting_cost'),
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total')
-    revenue = fields.Function(fields.Numeric(
+            currency='currency', digits='currency'),
+        'get_total')
+    revenue = fields.Function(Monetary(
             lazy_gettext('stock.msg_stock_reporting_revenue'),
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total')
-    profit = fields.Function(fields.Numeric(
+            currency='currency', digits='currency'),
+        'get_total')
+    profit = fields.Function(Monetary(
             lazy_gettext('stock.msg_stock_reporting_profit'),
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total')
-    margin = fields.Function(fields.Numeric(
+            currency='currency', digits='currency'),
+        'get_total')
+    margin = fields.Function(Monetary(
             lazy_gettext('stock.msg_stock_reporting_margin'),
-            digits=(14, 4),
-            depends=['currency_digits']),
+            digits=(14, 4)),
         'get_margin')
 
-    currency_digits = fields.Function(fields.Integer(
-            lazy_gettext('stock.msg_stock_reporting_currency_digits')),
-        'get_currency_digits')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency',
+            lazy_gettext('stock.msg_stock_reporting_currency')),
+        'get_currency')
 
     @classmethod
     def __setup__(cls):
@@ -542,6 +577,12 @@ class CategoryTree(ModelSQL, ModelView):
         categories = Category.browse(categories)
         return {c.id: c.name for c in categories}
 
+    @classmethod
+    def search_name(cls, name, clause):
+        pool = Pool()
+        Category = pool.get('product.category')
+        return [('id', 'in', Category.search([clause], query=True))]
+
     @classmethod
     def order_name(cls, tables):
         pool = Pool()
@@ -562,7 +603,6 @@ class CategoryTree(ModelSQL, ModelView):
     def get_total(cls, categories, names):
         pool = Pool()
         ReportingCategory = pool.get('stock.reporting.margin.category')
-        table = cls.__table__()
         reporting_category = ReportingCategory.__table__()
         cursor = Transaction().connection.cursor()
 
@@ -570,14 +610,9 @@ class CategoryTree(ModelSQL, ModelView):
                 ('parent', 'child_of', [c.id for c in categories]),
                 ])
         ids = [c.id for c in categories]
-        parents = {}
         reporting_categories = []
         for sub_ids in grouped_slice(ids):
             sub_ids = list(sub_ids)
-            where = reduce_ids(table.id, sub_ids)
-            cursor.execute(*table.select(table.id, table.parent, where=where))
-            parents.update(cursor)
-
             where = reduce_ids(reporting_category.id, sub_ids)
             cursor.execute(
                 *reporting_category.select(reporting_category.id, where=where))
@@ -586,31 +621,10 @@ class CategoryTree(ModelSQL, ModelView):
         result = {}
         reporting_categories = ReportingCategory.browse(reporting_categories)
         for name in names:
-            values = dict.fromkeys(ids, 0)
-            values.update(
-                (c.id, getattr(c, name)) for c in reporting_categories)
-            result[name] = cls._sum_tree(categories, values, parents)
-        return result
-
-    @classmethod
-    def _sum_tree(cls, categories, values, parents):
-        result = values.copy()
-        categories = set((c.id for c in categories))
-        leafs = categories - set(parents.values())
-        while leafs:
-            for category in leafs:
-                categories.remove(category)
-                parent = parents.get(category)
-                if parent in result:
-                    result[parent] += result[category]
-            next_leafs = set(categories)
-            for category in categories:
-                parent = parents.get(category)
-                if not parent:
-                    continue
-                if parent in next_leafs and parent in categories:
-                    next_leafs.remove(parent)
-            leafs = next_leafs
+            values = defaultdict(
+                Decimal,
+                {c.id: getattr(c, name) for c in reporting_categories})
+            result[name] = sum_tree(categories, values)
         return result
 
     def get_margin(self, name):
@@ -619,12 +633,12 @@ class CategoryTree(ModelSQL, ModelView):
             return (self.profit / self.revenue).quantize(
                 Decimal(1) / 10 ** digits[1])
 
-    def get_currency_digits(self, name):
+    def get_currency(self, name):
         pool = Pool()
         Company = pool.get('company.company')
         company = Transaction().context.get('company')
         if company:
-            return Company(company).currency.digits
+            return Company(company).currency.id
 
     @classmethod
     def view_attributes(cls):
